@@ -50,7 +50,10 @@
 #include "qwaylandinputdevice_p.h"
 #include "qwaylandinputmethodeventbuilder_p.h"
 #include "qwaylandwindow_p.h"
-#include "qwaylandxkb_p.h"
+
+#if QT_CONFIG(xkbcommon)
+#include <locale.h>
+#endif
 
 QT_BEGIN_NAMESPACE
 
@@ -120,7 +123,8 @@ void QWaylandTextInput::updateState(Qt::InputMethodQueries queries, uint32_t fla
     if (!QGuiApplication::focusWindow() || !QGuiApplication::focusWindow()->handle())
         return;
 
-    struct ::wl_surface *surface = static_cast<QWaylandWindow *>(QGuiApplication::focusWindow()->handle())->object();
+    auto *window = static_cast<QWaylandWindow *>(QGuiApplication::focusWindow()->handle());
+    auto *surface = window->wlSurface();
     if (!surface || (surface != m_surface))
         return;
 
@@ -158,8 +162,10 @@ void QWaylandTextInput::updateState(Qt::InputMethodQueries queries, uint32_t fla
 
     if (queries & Qt::ImCursorRectangle) {
         const QRect &cRect = event.value(Qt::ImCursorRectangle).toRect();
-        const QRect &tRect = QGuiApplication::inputMethod()->inputItemTransform().mapRect(cRect);
-        set_cursor_rectangle(tRect.x(), tRect.y(), tRect.width(), tRect.height());
+        const QRect &windowRect = QGuiApplication::inputMethod()->inputItemTransform().mapRect(cRect);
+        const QMargins margins = window->frameMargins();
+        const QRect &surfaceRect = windowRect.translated(margins.left(), margins.top());
+        set_cursor_rectangle(surfaceRect.x(), surfaceRect.y(), surfaceRect.width(), surfaceRect.height());
     }
 
     if (queries & Qt::ImPreferredLanguage) {
@@ -222,11 +228,11 @@ void QWaylandTextInput::zwp_text_input_v2_leave(uint32_t serial, ::wl_surface *s
 
 void QWaylandTextInput::zwp_text_input_v2_modifiers_map(wl_array *map)
 {
-    QList<QByteArray> modifiersMap = QByteArray::fromRawData(static_cast<const char*>(map->data), map->size).split('\0');
+    const QList<QByteArray> modifiersMap = QByteArray::fromRawData(static_cast<const char*>(map->data), map->size).split('\0');
 
     m_modifiersMap.clear();
 
-    Q_FOREACH (const QByteArray &modifier, modifiersMap) {
+    for (const QByteArray &modifier : modifiersMap) {
         if (modifier == "Shift")
             m_modifiersMap.append(Qt::ShiftModifier);
         else if (modifier == "Control")
@@ -315,6 +321,7 @@ void QWaylandTextInput::zwp_text_input_v2_delete_surrounding_text(uint32_t befor
 
 void QWaylandTextInput::zwp_text_input_v2_keysym(uint32_t time, uint32_t sym, uint32_t state, uint32_t modifiers)
 {
+#if QT_CONFIG(xkbcommon)
     if (m_resetCallback) {
         qCDebug(qLcQpaInputMethods()) << "discard keysym: reset not confirmed";
         return;
@@ -325,13 +332,18 @@ void QWaylandTextInput::zwp_text_input_v2_keysym(uint32_t time, uint32_t sym, ui
 
     Qt::KeyboardModifiers qtModifiers = modifiersToQtModifiers(modifiers);
 
-    QEvent::Type type = QWaylandXkb::toQtEventType(state);
-    QString text;
-    int qtkey;
-    std::tie(qtkey, text) = QWaylandXkb::keysymToQtKey(sym, qtModifiers);
+    QEvent::Type type = state == WL_KEYBOARD_KEY_STATE_PRESSED ? QEvent::KeyPress : QEvent::KeyRelease;
+    QString text = QXkbCommon::lookupStringNoKeysymTransformations(sym);
+    int qtkey = QXkbCommon::keysymToQtKey(sym, qtModifiers);
 
     QWindowSystemInterface::handleKeyEvent(QGuiApplication::focusWindow(),
                                            time, type, qtkey, qtModifiers, text);
+#else
+    Q_UNUSED(time);
+    Q_UNUSED(sym);
+    Q_UNUSED(state);
+    Q_UNUSED(modifiers);
+#endif
 }
 
 void QWaylandTextInput::zwp_text_input_v2_language(const QString &language)
@@ -375,8 +387,10 @@ void QWaylandTextInput::zwp_text_input_v2_input_method_changed(uint32_t serial,
 Qt::KeyboardModifiers QWaylandTextInput::modifiersToQtModifiers(uint32_t modifiers)
 {
     Qt::KeyboardModifiers ret = Qt::NoModifier;
-    for (int i = 0; modifiers >>= 1; ++i) {
-        ret |= m_modifiersMap[i];
+    for (int i = 0; i < m_modifiersMap.size(); ++i) {
+        if (modifiers & (1 << i)) {
+            ret |= m_modifiersMap[i];
+        }
     }
     return ret;
 }
@@ -398,6 +412,10 @@ bool QWaylandInputContext::isValid() const
 void QWaylandInputContext::reset()
 {
     qCDebug(qLcQpaInputMethods) << Q_FUNC_INFO;
+#if QT_CONFIG(xkbcommon)
+    if (m_composeState)
+        xkb_compose_state_reset(m_composeState);
+#endif
 
     QPlatformInputContext::reset();
 
@@ -423,7 +441,7 @@ static ::wl_surface *surfaceForWindow(QWindow *window)
         return nullptr;
 
     auto *waylandWindow = static_cast<QWaylandWindow *>(window->handle());
-    return waylandWindow->wl_surface::object();
+    return waylandWindow->wlSurface();
 }
 
 void QWaylandInputContext::update(Qt::InputMethodQueries queries)
@@ -518,9 +536,14 @@ Qt::LayoutDirection QWaylandInputContext::inputDirection() const
     return textInput()->inputDirection();
 }
 
-void QWaylandInputContext::setFocusObject(QObject *)
+void QWaylandInputContext::setFocusObject(QObject *object)
 {
     qCDebug(qLcQpaInputMethods) << Q_FUNC_INFO;
+#if QT_CONFIG(xkbcommon)
+    m_focusObject = object;
+#else
+    Q_UNUSED(object);
+#endif
 
     if (!textInput())
         return;
@@ -529,7 +552,7 @@ void QWaylandInputContext::setFocusObject(QObject *)
 
     if (mCurrentWindow && mCurrentWindow->handle()) {
         if (mCurrentWindow.data() != window || !inputMethodAccepted()) {
-            struct ::wl_surface *surface = static_cast<QWaylandWindow *>(mCurrentWindow->handle())->object();
+            auto *surface = static_cast<QWaylandWindow *>(mCurrentWindow->handle())->wlSurface();
             if (surface)
                 textInput()->disable(surface);
             mCurrentWindow.clear();
@@ -538,7 +561,7 @@ void QWaylandInputContext::setFocusObject(QObject *)
 
     if (window && window->handle() && inputMethodAccepted()) {
         if (mCurrentWindow.data() != window) {
-            struct ::wl_surface *surface = static_cast<QWaylandWindow *>(window->handle())->object();
+            auto *surface = static_cast<QWaylandWindow *>(window->handle())->wlSurface();
             if (surface) {
                 textInput()->enable(surface);
                 mCurrentWindow = window;
@@ -553,6 +576,92 @@ QWaylandTextInput *QWaylandInputContext::textInput() const
     return mDisplay->defaultInputDevice()->textInput();
 }
 
+#if QT_CONFIG(xkbcommon)
+
+void QWaylandInputContext::ensureInitialized()
+{
+    if (m_initialized)
+        return;
+
+    if (!m_XkbContext) {
+        qCWarning(qLcQpaInputMethods) << "error: xkb context has not been set on" << metaObject()->className();
+        return;
+    }
+
+    m_initialized = true;
+    const char *locale = setlocale(LC_CTYPE, "");
+    if (!locale)
+        locale = setlocale(LC_CTYPE, nullptr);
+    qCDebug(qLcQpaInputMethods) << "detected locale (LC_CTYPE):" << locale;
+
+    m_composeTable = xkb_compose_table_new_from_locale(m_XkbContext, locale, XKB_COMPOSE_COMPILE_NO_FLAGS);
+    if (m_composeTable)
+        m_composeState = xkb_compose_state_new(m_composeTable, XKB_COMPOSE_STATE_NO_FLAGS);
+
+    if (!m_composeTable) {
+        qCWarning(qLcQpaInputMethods, "failed to create compose table");
+        return;
+    }
+    if (!m_composeState) {
+        qCWarning(qLcQpaInputMethods, "failed to create compose state");
+        return;
+    }
+}
+
+bool QWaylandInputContext::filterEvent(const QEvent *event)
+{
+    auto keyEvent = static_cast<const QKeyEvent *>(event);
+    if (keyEvent->type() != QEvent::KeyPress)
+        return false;
+
+    if (!inputMethodAccepted())
+        return false;
+
+    // lazy initialization - we don't want to do this on an app startup
+    ensureInitialized();
+
+    if (!m_composeTable || !m_composeState)
+        return false;
+
+    xkb_compose_state_feed(m_composeState, keyEvent->nativeVirtualKey());
+
+    switch (xkb_compose_state_get_status(m_composeState)) {
+    case XKB_COMPOSE_COMPOSING:
+        return true;
+    case XKB_COMPOSE_CANCELLED:
+        reset();
+        return false;
+    case XKB_COMPOSE_COMPOSED:
+    {
+        const int size = xkb_compose_state_get_utf8(m_composeState, nullptr, 0);
+        QVarLengthArray<char, 32> buffer(size + 1);
+        xkb_compose_state_get_utf8(m_composeState, buffer.data(), buffer.size());
+        QString composedText = QString::fromUtf8(buffer.constData());
+
+        QInputMethodEvent event;
+        event.setCommitString(composedText);
+
+        if (!m_focusObject && qApp)
+            m_focusObject = qApp->focusObject();
+
+        if (m_focusObject)
+            QCoreApplication::sendEvent(m_focusObject, &event);
+        else
+            qCWarning(qLcQpaInputMethods, "no focus object");
+
+        reset();
+        return true;
+    }
+    case XKB_COMPOSE_NOTHING:
+        return false;
+    default:
+        Q_UNREACHABLE();
+        return false;
+    }
+}
+
+#endif
+
 }
 
 QT_END_NAMESPACE
