@@ -6,7 +6,7 @@
 **
 ** This file is part of the QtWaylandCompositor module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:GPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
@@ -15,24 +15,14 @@
 ** and conditions see https://www.qt.io/terms-conditions. For further
 ** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** General Public License version 3 or (at your option) any later version
+** approved by the KDE Free Qt Foundation. The licenses are as published by
+** the Free Software Foundation and appearing in the file LICENSE.GPL3
 ** included in the packaging of this file. Please review the following
 ** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -44,6 +34,7 @@
 #if QT_CONFIG(wayland_compositor_quick)
 #include "qwaylandwlshellintegration_p.h"
 #endif
+#include <QtWaylandCompositor/private/qwaylandutils_p.h>
 
 #include <QtWaylandCompositor/QWaylandCompositor>
 #include <QtWaylandCompositor/QWaylandView>
@@ -254,6 +245,7 @@ void QWaylandWlShellSurfacePrivate::shell_surface_set_class(Resource *resource,
 /*!
  * \qmltype WlShell
  * \inqmlmodule QtWayland.Compositor
+ * \instantiates QWaylandWlShell
  * \since 5.8
  * \brief Provides an extension for desktop-style user interfaces.
  *
@@ -266,15 +258,16 @@ void QWaylandWlShellSurfacePrivate::shell_surface_set_class(Resource *resource,
  * To provide the functionality of the shell extension in a compositor, create
  * an instance of the WlShell component and add it to the list of extensions
  * supported by the compositor:
- * \code
- * import QtWayland.Compositor 1.0
+ *
+ * \qml \QtMinorVersion
+ * import QtWayland.Compositor 1.\1
  *
  * WaylandCompositor {
  *     WlShell {
  *         // ...
  *     }
  * }
- * \endcode
+ * \endqml
  */
 
 /*!
@@ -330,7 +323,7 @@ QList<QWaylandWlShellSurface *> QWaylandWlShell::shellSurfacesForClient(QWayland
 {
     Q_D(const QWaylandWlShell);
     QList<QWaylandWlShellSurface *> surfsForClient;
-    Q_FOREACH (QWaylandWlShellSurface *shellSurface, d->m_shellSurfaces) {
+    for (QWaylandWlShellSurface *shellSurface : d->m_shellSurfaces) {
         if (shellSurface->surface() && shellSurface->surface()->client() == client)
             surfsForClient.append(shellSurface);
     }
@@ -341,7 +334,7 @@ QList<QWaylandWlShellSurface *> QWaylandWlShell::mappedPopups() const
 {
     Q_D(const QWaylandWlShell);
     QList<QWaylandWlShellSurface *> popupSurfaces;
-    Q_FOREACH (QWaylandWlShellSurface *shellSurface, d->m_shellSurfaces) {
+    for (QWaylandWlShellSurface *shellSurface : d->m_shellSurfaces) {
         if (shellSurface->windowType() == Qt::WindowType::Popup
                 && shellSurface->surface() && shellSurface->surface()->hasContent()) {
             popupSurfaces.append(shellSurface);
@@ -353,7 +346,7 @@ QList<QWaylandWlShellSurface *> QWaylandWlShell::mappedPopups() const
 QWaylandClient *QWaylandWlShell::popupClient() const
 {
     Q_D(const QWaylandWlShell);
-    Q_FOREACH (QWaylandWlShellSurface *shellSurface, d->m_shellSurfaces) {
+    for (QWaylandWlShellSurface *shellSurface : d->m_shellSurfaces) {
         if (shellSurface->windowType() == Qt::WindowType::Popup
                 && shellSurface->surface() && shellSurface->surface()->hasContent()) {
             return shellSurface->surface()->client();
@@ -364,7 +357,8 @@ QWaylandClient *QWaylandWlShell::popupClient() const
 
 void QWaylandWlShell::closeAllPopups()
 {
-    Q_FOREACH (QWaylandWlShellSurface* shellSurface, mappedPopups())
+    const auto mapped = mappedPopups();
+    for (QWaylandWlShellSurface *shellSurface : mapped)
         shellSurface->sendPopupDone();
 }
 
@@ -421,6 +415,7 @@ QByteArray QWaylandWlShell::interfaceName()
 /*!
  * \qmltype WlShellSurface
  * \inqmlmodule QtWayland.Compositor
+ * \instantiates QWaylandWlShellSurface
  * \since 5.8
  * \brief Provides a \c wl_shell_surface that offers desktop-style compositor-specific features to a surface.
  *
@@ -464,7 +459,8 @@ QWaylandWlShellSurface::QWaylandWlShellSurface(QWaylandWlShell *shell, QWaylandS
 QWaylandWlShellSurface::~QWaylandWlShellSurface()
 {
     Q_D(QWaylandWlShellSurface);
-    QWaylandWlShellPrivate::get(d->m_shell)->unregisterShellSurface(this);
+    if (d->m_shell)
+        QWaylandWlShellPrivate::get(d->m_shell)->unregisterShellSurface(this);
 }
 
 /*!
@@ -699,9 +695,8 @@ void QWaylandWlShellSurface::ping()
  */
 QWaylandWlShellSurface *QWaylandWlShellSurface::fromResource(wl_resource *resource)
 {
-    QWaylandWlShellSurfacePrivate::Resource *res = QWaylandWlShellSurfacePrivate::Resource::fromResource(resource);
-    if (res)
-        return static_cast<QWaylandWlShellSurfacePrivate *>(res->shell_surface_object)->q_func();
+    if (auto p = QtWayland::fromResource<QWaylandWlShellSurfacePrivate *>(resource))
+        return p->q_func();
     return nullptr;
 }
 
